package com.tencent.subscribe.demo;

import com.tencent.subscribe.SubscribeDataProto;
import org.apache.commons.cli.*;
import org.apache.kafka.clients.CommonClientConfigs;
import org.apache.kafka.clients.consumer.*;
import org.apache.kafka.common.TopicPartition;
import org.apache.kafka.common.config.SaslConfigs;
import org.apache.kafka.common.header.Header;

import javax.xml.bind.DatatypeConverter;
import java.io.ByteArrayOutputStream;
import java.time.Duration;
import java.time.Instant;
import java.time.LocalDateTime;
import java.util.*;

public class SubscribeFromKafka {
    private static Map<Integer, partitionMsgConsumer> partitionMsgConsumers = new HashMap<>();


    /**
     * 本Demo基本执行流程为：
     * 1. 创建 Kafka 消费者，在一个无限循环中调用消费者的 poll 方法不停消费消息；
     * 2. 依次遍历每条消息，根据消息中的分区找到分区对应的 partitionMsgConsumer 对象，由该对象对消息进行处理；
     * 3. partitionMsgConsumer 检查消息头的序列号，如果发现重复消息则丢弃，如果发现消息丢失则终止运行；
     * 4. partitionMsgConsumer 将原始消息反序列化为 Envelope 结构，如果 Envelope.Data 存在分段则进行拼接；
     * 5. 将 Envelope.Data 用 Protobuf 解码为 Entries；
     * 6. 遍历 Entries 中的每一个 Entry，将数据打印出来。
     **/
    public static void main(final String[] args) {
        Option option_brokers = Option.builder()
                .required(true)
                .desc("The brokers option")
                .longOpt("brokers")
                .hasArg()
                .optionalArg(false)
                .build();
        Option option_topic = Option.builder()
                .required(true)
                .desc("The topic option")
                .longOpt("topic")
                .hasArg()
                .optionalArg(false)
                .build();
        Option option_group = Option.builder()
                .required(true)
                .desc("The group option")
                .longOpt("group")
                .hasArg()
                .optionalArg(false)
                .build();
        Option option_user = Option.builder()
                .required(false)
                .desc("The user option")
                .longOpt("user")
                .hasArg()
                .optionalArg(false)
                .build();
        Option option_password = Option.builder()
                .required(false)
                .desc("The password option")
                .longOpt("password")
                .hasArg()
                .optionalArg(false)
                .build();
        Option option_trans = Option.builder()
                .required(false)
                .desc("The trans2sql option")
                .longOpt("trans2sql")
                .build();

        Options options = new Options()
                .addOption(option_brokers)
                .addOption(option_topic)
                .addOption(option_user)
                .addOption(option_password)
                .addOption(option_group)
                .addOption(option_trans);

        CommandLineParser parser = new DefaultParser();

        try {
            CommandLine commandLine = parser.parse(options, args);

            run(
                    commandLine.getOptionValue("brokers"),
                    commandLine.getOptionValue("topic"),
                    commandLine.getOptionValue("group"),
                    commandLine.getOptionValue("user"),
                    commandLine.getOptionValue("password"),
                    commandLine.hasOption("trans2sql")
            );

        } catch (Exception exception) {
            System.out.print("error: ");
            System.out.println(exception.getMessage());
        }
    }

    private static void run(
            String brokers,
            String topic,
            String group,
            String user,
            String password,
            boolean trans2sql) throws Exception {
        Properties props = new Properties();
        props.put(ConsumerConfig.BOOTSTRAP_SERVERS_CONFIG, brokers);
        props.put(ConsumerConfig.GROUP_ID_CONFIG, group);
        props.put(ConsumerConfig.CLIENT_ID_CONFIG, group + "_" + Instant.now().getEpochSecond());
        props.put(ConsumerConfig.ENABLE_AUTO_COMMIT_CONFIG, false);
        props.put(ConsumerConfig.ISOLATION_LEVEL_CONFIG, "read_committed");
        props.put(ConsumerConfig.AUTO_OFFSET_RESET_CONFIG, "earliest");
        props.put(ConsumerConfig.KEY_DESERIALIZER_CLASS_CONFIG, "org.apache.kafka.common.serialization.StringDeserializer");
        props.put(ConsumerConfig.VALUE_DESERIALIZER_CLASS_CONFIG, "org.apache.kafka.common.serialization.ByteArrayDeserializer");

        if (null != user && !user.equals("")) {
            props.put(CommonClientConfigs.SECURITY_PROTOCOL_CONFIG, "SASL_PLAINTEXT");
            props.put(SaslConfigs.SASL_MECHANISM, "SCRAM-SHA-512");
            props.put(SaslConfigs.SASL_JAAS_CONFIG,
                    "org.apache.kafka.common.security.scram.ScramLoginModule required " +
                            "  username=\"" + user + "\"" +
                            "  password=\"" + password + "\";");
        }

        KafkaConsumer<String, byte[]> consumer = new KafkaConsumer<>(props);
        consumer.subscribe(Collections.singletonList(topic), new partitionSubLis());
        try {
            while (true) {
                ConsumerRecords<String, byte[]> records = consumer.poll(Duration.ofSeconds(1));
                for (ConsumerRecord<String, byte[]> record : records) {
                    partitionMsgConsumer pc = partitionMsgConsumers.get(record.partition());
                    if (null == pc) {
                        throw new IllegalStateException("no such partition consumer " + record.partition());
                    }

                    pc.consume(consumer, record, trans2sql);
                }
            }
        } finally {
            consumer.close();
        }
    }

    private final static String partitionSeqKey = "ps";
    private final static String partitionSeq2Key = "ps2";
    private final static String CHECKPOINT = "Checkpoint";
    private final static String HEARTBEAT = "Heartbeat";

    private static long getPartitionSeq(ConsumerRecord<String, byte[]> msg) {
        Iterator<Header> it = msg.headers().headers(partitionSeqKey).iterator();
        if (it.hasNext()) {
            Header h = it.next();
            return Long.parseUnsignedLong(new String(h.value()));
        }
        throw new IllegalStateException(partitionSeqKey + " does not exists");
    }

    private static long getPartitionSeq2(ConsumerRecord<String, byte[]> msg) {
        Iterator<Header> it = msg.headers().headers(partitionSeq2Key).iterator();
        if (it.hasNext()) {
            Header h = it.next();
            return Long.parseUnsignedLong(new String(h.value()));
        }
        return -1;
    }

    private static boolean isCheckpoint(ConsumerRecord<String, byte[]> msg) {
        Iterator<Header> it = msg.headers().headers(CHECKPOINT).iterator();
        return it.hasNext();
    }

    private static boolean isHeartbeat(ConsumerRecord<String, byte[]> msg) {
        Iterator<Header> it = msg.headers().headers(HEARTBEAT).iterator();
        return it.hasNext();
    }

    /**
     * 如果传入trans2sql参数，将Entry里的数据解析为sql语句并打印出来，否则直接将Entry打印出来。
     * 您可以在此函数基础上开发满足自己需要的数据处理逻辑。
     **/
    private static void onEntry(int partitionNum, long offset, long partitionSeq, SubscribeDataProto.Entry entry, boolean trans2sql) throws Exception {
        if (trans2sql) {
            trans2SQL(partitionNum, offset, partitionSeq, entry);
        } else {
            System.out.println("partition:" + partitionNum + ", offset:" + offset + ", partitionSeq:" + partitionSeq
                    + ", " + entry.toString());
        }
    }

    private static void trans2SQL(int partitionNum, long offset, long partitionSeq, SubscribeDataProto.Entry entry) throws Exception {
        SubscribeDataProto.Header header = entry.getHeader();
        System.out.println(
                String.format("-->[partition: %d, offset: %d, partitionSeq: %d] [%s:%d], happenedAt: %s",
                        partitionNum,
                        offset,
                        partitionSeq,
                        header.getFileName(), header.getPosition(),
                        LocalDateTime.ofInstant(Instant.ofEpochSecond(header.getTimestamp()),
                                TimeZone.getDefault().toZoneId())));

        switch (entry.getHeader().getMessageType()) {
            case HEARTBEAT:
                System.out.println("heartbeat: " + entry.getEvent().getHeartbeatEvent().getEpoch());
                break;
            case CHECKPOINT:
                SubscribeDataProto.CheckpointEvent evt = entry.getEvent().getCheckpointEvent();
                System.out.println("Checkpoint: gtid=" + evt.getSyncedGtid()
                        + ", file=" + evt.getFileName() + ", position=" + evt.getPosition());
                break;
            case BEGIN:
                System.out.println("BEGIN");
                break;
            case COMMIT:
                System.out.println("COMMIT");
                break;
            case ROLLBACK:
                System.out.println("ROLLBACK");
                break;
            case DDL:
                System.out.println(entry.getEvent().getDdlEvent().getSql());
                break;
            case DML:
                switch (entry.getEvent().getDmlEvent().getDmlEventType()) {
                    case INSERT:
                        transInsert(entry);
                        break;
                    case DELETE:
                        transDelete(entry);
                        break;
                    case UPDATE:
                        transUpdate(entry);
                        break;
                    default:
                        throw new IllegalStateException("unsupported dml event type");
                }
                break;
            default:
                throw new IllegalStateException("unsupported message type");
        }
    }

    private static void transInsert(SubscribeDataProto.Entry entry) throws Exception {
        SubscribeDataProto.DMLEvent dmlEvt = entry.getEvent().getDmlEvent();
        System.out.print(
                String.format(
                        "INSERT INTO `%s`.`%s` VALUES ",
                        entry.getHeader().getSchemaName(),
                        entry.getHeader().getTableName()));
        String rowSep = "";
        for (SubscribeDataProto.RowChange row : dmlEvt.getRowsList()) {
            System.out.print(rowSep);
            System.out.print('(');
            String sep = "";
            for (SubscribeDataProto.Data col : row.getNewColumnsList()) {
                System.out.print(sep);
                System.out.print(decode(col));
                sep = ", ";
            }
            System.out.print(')');
            rowSep = ", ";
        }

        System.out.println();
    }

    private static void transDelete(SubscribeDataProto.Entry entry) throws Exception {
        SubscribeDataProto.DMLEvent dmlEvt = entry.getEvent().getDmlEvent();
        String header = String.format(
                "DELETE FROM `%s`.`%s` WHERE ",
                entry.getHeader().getSchemaName(),
                entry.getHeader().getTableName());
        for (SubscribeDataProto.RowChange row : dmlEvt.getRowsList()) {
            System.out.print(header);
            String sep = "";
            for (int i = 0; i < row.getOldColumnsCount(); i++) {
                SubscribeDataProto.Data col = row.getOldColumns(i);
                if (col.getDataType() == SubscribeDataProto.DataType.NA) {
                    continue;
                }

                System.out.print(sep);
                SubscribeDataProto.Column colDef = dmlEvt.getColumns(i);
                System.out.print(String.format("`%s`", colDef.getName()));
                if (col.getDataType() == SubscribeDataProto.DataType.NIL) {
                    System.out.print(" IS NULL");
                } else {
                    System.out.print(String.format(" = %s", decode(col)));
                }
                sep = " AND ";
            }
            System.out.println(" LIMIT 1");
        }
    }

    private static void transUpdate(SubscribeDataProto.Entry entry) throws Exception {
        SubscribeDataProto.DMLEvent dmlEvt = entry.getEvent().getDmlEvent();
        String header = String.format(
                "UPDATE `%s`.`%s` SET ",
                entry.getHeader().getSchemaName(),
                entry.getHeader().getTableName());
        for (SubscribeDataProto.RowChange row : dmlEvt.getRowsList()) {
            System.out.print(header);
            String sep = "";
            for (int i = 0; i < row.getNewColumnsCount(); i++) {
                SubscribeDataProto.Data col = row.getNewColumns(i);
                if (col.getDataType() == SubscribeDataProto.DataType.NA) {
                    continue;
                }

                System.out.print(sep);
                SubscribeDataProto.Column colDef = dmlEvt.getColumns(i);
                System.out.print(String.format("`%s`", colDef.getName()));
                if (col.getDataType() == SubscribeDataProto.DataType.NIL) {
                    System.out.print(" IS NULL");
                } else {
                    System.out.print(String.format(" = %s", decode(col)));
                }
                sep = ", ";
            }

            System.out.print(" WHERE ");
            sep = "";
            for (int i = 0; i < row.getOldColumnsCount(); i++) {
                SubscribeDataProto.Data col = row.getOldColumns(i);
                if (col.getDataType() == SubscribeDataProto.DataType.NA) {
                    continue;
                }

                System.out.print(sep);
                SubscribeDataProto.Column colDef = dmlEvt.getColumns(i);
                System.out.print(String.format("`%s`", colDef.getName()));
                if (col.getDataType() == SubscribeDataProto.DataType.NIL) {
                    System.out.print(" IS NULL");
                } else {
                    System.out.print(String.format(" = %s", decode(col)));
                }
                sep = " AND ";
            }
            System.out.println(" LIMIT 1");
        }
    }

    private static String decode(SubscribeDataProto.Data data) throws Exception {
        switch (data.getDataType()) {
            case INT8:
            case INT16:
            case INT32:
            case INT64:
            case UINT8:
            case UINT16:
            case UINT32:
            case UINT64:
            case DECIMAL:
            case FLOAT32:
            case FLOAT64:
                return data.getSv();
            case STRING:
                return "_binary'" + new String(data.getBv().toByteArray()) + "'";
            case BYTES:
                return "x'" + DatatypeConverter.printHexBinary(data.getBv().toByteArray()) + "'";
            case NA:
                return "DEFAULT";
            case NIL:
                return "NULL";
            default:
                throw new IllegalStateException("unsupported data type");
        }
    }

    private static class partitionMsgConsumer {
        ByteArrayOutputStream completeMsg = new ByteArrayOutputStream();
        long lastPS = 0;
        long lastPS2 = 0;
        TopicPartition tp;

        private partitionMsgConsumer(TopicPartition tp) {
            this.tp = tp;
        }

        private void consume(KafkaConsumer<String, byte[]> consumer, ConsumerRecord<String, byte[]> record, boolean trans2sql) throws Exception {
            // 每个消息的消息头部都有一个编号记做ps，这个编号在同一个分区内正常情况下保证单调递增且连续
            // ps2是ps的升级版，2022.11.1之后启动的订阅任务包含此属性，可以最大限度地解决Kafka消息重复投递的问题，
            // ps解决的是Kafka分区中消息重复投递(Kafka自身机制导致)的问题，ps2在ps的基础上还可以解决后端任务重启后
            // 重复解析部分binlog导致的数据重复投递问题。
            long ps2 = getPartitionSeq2(record);
            long validPs;
            if (ps2 != -1) {
                if (lastPS2 > 0) {
                    if (ps2 < lastPS2) {
                        System.out.println("Duplicated message: " + record);
                        return;
                    } else if (ps2 == lastPS2) { // 可能是重复消息也可能是checkpoint或heartbeat消息
                        if (!isCheckpoint(record) && !isHeartbeat(record)) {
                            System.out.println("Duplicated message: " + record);
                            return;
                        }
                    } else if (ps2 != lastPS2 + 1) {
                        // 在极端情况下，kafka也可能出现丢失消息的情况（kafka自身机制的原因，概率极小），此时partitionSeq不连续，
                        // 如果对数据一致性要求较高，需要停止消费数据，然后在控制台重置订阅任务再重新配置。
                        // 如果对数据一致性要求不高，可以接受个别消息丢失，那么跳过此条件判断即可。
                        throw new IllegalStateException("the partition seq2 is discontinuous, last: " + lastPS2 + ", current: " + ps2);
                    }
                }
                lastPS2 = ps2;
                validPs = ps2;
            } else { // 老的订阅任务没有ps2，会走到这里
                long ps = getPartitionSeq(record);
                if (ps <= lastPS) {
                    System.out.println("Duplicated message: " + record);
                    return;
                } else if (ps != lastPS + 1 && lastPS != 0) {
                    throw new IllegalStateException("the partition seq is discontinuous, last: " + lastPS + ", current: " + ps);
                }
                lastPS = ps;
                validPs = ps;
            }

            // 将kafka消息的Value值转换为Envelope
            SubscribeDataProto.Envelope envelope = SubscribeDataProto.Envelope.parseFrom(record.value());
            if (1 != envelope.getVersion()) {
                throw new IllegalStateException(String.format("unsupported version: %d", envelope.getVersion()));
            }

            //Envelope的Data字段对应Entries二进制序列，Entries包含一个或多个Event。在封装消费的过程中，
            //当一个Event的大小超过单个kafka消息大小限制时，Entries过大，需要将Entries二进制序列分割为多段，
            //每段装入一个Envelope，total和index字段分别记录总段数和分段序号，所以这里需要将分段的Data做拼接。
            if (0 == envelope.getIndex()) {
                completeMsg = new ByteArrayOutputStream();
                envelope.getData().writeTo(completeMsg);
            } else {
                envelope.getData().writeTo(completeMsg);
            }

            if (envelope.getIndex() < envelope.getTotal() - 1) {
                return;
            }
            // 将Envelope.Data反序列化为Entries
            SubscribeDataProto.Entries entries = SubscribeDataProto.Entries.parseFrom(completeMsg.toByteArray());
            // 遍历Entries中的Entry，在onEntry函数中处理消费到的每个Entry
            for (SubscribeDataProto.Entry entry : entries.getItemsList()) {
                onEntry(record.partition(), record.offset(), validPs, entry, trans2sql);
            }
            // 当消费到CHECKPOINT消息时，做一次提交。CHECKPOIN消息在entries中单独存在，所以取Items第一个值判断即可。
            if (entries.getItemsCount() > 0
                    && entries.getItems(0).getHeader().getMessageType() == SubscribeDataProto.MessageType.CHECKPOINT) {
                consumer.commitAsync(Collections.singletonMap(tp, new OffsetAndMetadata(record.offset() + 1)), null);
            }
        }
    }

    private static class partitionSubLis implements ConsumerRebalanceListener {
        public void onPartitionsAssigned(Collection<TopicPartition> partitions) {
            for (TopicPartition p : partitions) {
                partitionMsgConsumers.computeIfAbsent(p.partition(), k -> new partitionMsgConsumer(p));
            }
        }

        public void onPartitionsRevoked(Collection<TopicPartition> partitions) {
            for (TopicPartition p : partitions) {
                partitionMsgConsumers.remove(p.partition());
            }
        }
    }
}
